%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This script illustrate multi-dimensional integration methods using a
%%% 2-dimensional example. The rotations considered are highly inspired by
%%% Jckel, P. (2005), "A note on multivariate Gauss-Hermite quadrature."
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% By Thomas H. Jrgensen, tjo@econ.ku.dk, 2013.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear all; clc
num = 8;                    % # of quadrature points
S = 10000;                    % # of Monte Carlo draws
func = @(x1,x2) sin(x1.*x2);     % Function, f: R^2->R to find expected value of 

% Specifying the moments of the bivariate normal distributed x1 and x2
se1 = 1;
se2 = 1;
mu1 = 0;
mu2 = 0;
mu = [mu1,mu2];
rho = .5;   % Correlation
Omega = [se1^2       rho*se1*se2
         rho*se1*se2  se2^2];

%%% The sqrt of the covarinace matrix is constructed three different ways
% 1. Cholesky decomposition
L = chol(Omega,'lower');

% 2. Eigenvalues
[V,D] = eig(Omega);
R = V * sqrt(D);

% 3. Rotation weights
a = (sqrt(1+rho) + sqrt(1-rho))/2;
b = (sqrt(1+rho) - sqrt(1-rho))/2;
% NOTE: This method seem to be what R = sqrtm(Omega); produces!

%%%%%%%%%%%%%%
% Quadrature %
[nodes, weights] = GaussHermiteNodes(num);      % The GaussHermite nodes and weights

% By loop, since easier to implement in C
Approx1 = 0.0;  Approx2 = 0.0;  Approx3 = 0.0;
for i=1:num
    for j=1:num
        nodes1 = [mu1 ; mu2] + L * [nodes(i) ; nodes(j)];
        Approx1 = Approx1 + weights(i)*weights(j)*func(nodes1(1),nodes1(2));
        
        nodes2 = [mu1 ; mu2] + R * [nodes(i) ; nodes(j)];
        Approx2 = Approx2 + weights(i)*weights(j)*func(nodes2(1),nodes2(2));
        
        nodes3 = [mu1 + se1*(a*nodes(i) + b*nodes(j))
                  mu2 + se2*(b*nodes(i) + a*nodes(j))];
        Approx3 = Approx3 + weights(i)*weights(j)*func(nodes3(1),nodes3(2));    
    end
end

%%%%%%%%%%%%%%%
% Monte Carlo %
rng(1270);          % Seed
Rand = randn(S,2);  % Draws

% 1. Cholesky
nodes4 = ones(S,1)*[mu1,mu2] + Rand*L'; 
Approx4 = mean( func(nodes4(:,1),nodes4(:,2)) );

% 2. Eigenvalues
nodes5 = ones(S,1)*[mu1,mu2] + Rand*R'; 
Approx5 = mean( func(nodes5(:,1),nodes5(:,2)) );

% 3. Rotation
nodes6 = [mu1 + se1*(a*Rand(:,1) + b*Rand(:,2)),...
          mu2 + se2*(b*Rand(:,1) + a*Rand(:,2))];
Approx6 = mean( func(nodes6(:,1),nodes6(:,2)) );

%%%%%%%%%%%%%%%%
% Sparse Grids %
degree = 5;%=round((num-1)/1.5)     % degree of polynomial to approximate --> accurate at 2*degree-1. 
[nodes_spa, weights] = SparseGridsNodes('GQN', 2, degree);
num_spa = size(nodes_spa,1);

nodes7 = ones(num_spa,1)*mu + nodes_spa*L';
nodes8 = ones(num_spa,1)*mu + nodes_spa*R';
nodes9 = [mu1 + se1*(a*nodes_spa(:,1) + b*nodes_spa(:,2)),...
          mu2 + se2*(b*nodes_spa(:,1) + a*nodes_spa(:,2))];
      
Approx7 = func(nodes7(:,1),nodes7(:,2))'*weights;   
Approx8 = func(nodes8(:,1),nodes8(:,2))'*weights;   
Approx9 = func(nodes9(:,1),nodes9(:,2))'*weights;   

%%%%%%%%%%%%%%%%
% Monomial     %
[ ~, x, weights ] = MonomialNodes( 2);
nodes_mo = x';
num_mo = size(nodes_mo,1);

nodes10 = ones(num_mo,1)*mu + nodes_mo*L';
nodes11 = ones(num_mo,1)*mu + nodes_mo*R';
nodes12 = [mu1 + se1*(a*nodes_mo(:,1) + b*nodes_mo(:,2)),...
          mu2 + se2*(b*nodes_mo(:,1) + a*nodes_mo(:,2))];
      
Approx10 = func(nodes10(:,1),nodes10(:,2))'*weights;   
Approx11 = func(nodes11(:,1),nodes11(:,2))'*weights;   
Approx12 = func(nodes12(:,1),nodes12(:,2))'*weights;  

%%%%%%%%%%%%%%%%%
% Print Results %
fprintf('\n             Cholesky  Eigenvals Rotation \n');
fprintf('Quadrature   %1.7f %1.7f %1.7f\nMonte Carlo  %1.7f %1.7f %1.7f\nSparse Grids %1.7f %1.7f %1.7f\nMonomial     %1.7f %1.7f %1.7f\n\n',Approx1,Approx2,Approx3,Approx4,Approx5,Approx6,Approx7,Approx8,Approx9,Approx10,Approx11,Approx12);

%% Plot the nodes
% Construct nodes regarding quadrature in vectorized form for plots
MATnodes = nodes*ones(1,num);
nodes1 = ones(num^2,1)*[mu1,mu2] + reshape([MATnodes , MATnodes'], num^2, 2) * L' ;
nodes2 = ones(num^2,1)*[mu1,mu2] + reshape([MATnodes , MATnodes'], num^2, 2) * R' ;
nodes3 = reshape([mu1 + se1*(a*MATnodes + b*MATnodes'),...
                  mu2 + se2*(b*MATnodes + a*MATnodes')], num^2, 2); 
% Domain used for plotting
dom = [-8,8];               

figure(1)
subplot(2,3,1)
scatter(nodes1(:,1),nodes1(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('Cholesky')
ylabel(['Quadrature, ',num2str(num),'^2 = ',num2str(num^2),' evaluations']); xlabel('')

subplot(2,3,2)
scatter(nodes2(:,1),nodes2(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('Eigenvalues')
ylabel(''); xlabel('')

subplot(2,3,3)
scatter(nodes3(:,1),nodes3(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('45^o Rotation')
ylabel(''); xlabel('')

subplot(2,3,4)
scatter(nodes7(:,1),nodes7(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(['Sparse Grids, ',num2str(num_spa),' evaluations']); xlabel('')

subplot(2,3,5)
scatter(nodes8(:,1),nodes8(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(''); xlabel('')

subplot(2,3,6)
scatter(nodes9(:,1),nodes9(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(''); xlabel('')

% FIGURE 2
figure(2)
subplot(2,3,1)
scatter(nodes10(:,1),nodes10(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('Cholesky')
ylabel(['Monomial, ',num2str(num_mo),' evaluations']); xlabel('')

subplot(2,3,2)
scatter(nodes11(:,1),nodes11(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('Eigenvalues')
ylabel(''); xlabel('')

subplot(2,3,3)
scatter(nodes12(:,1),nodes12(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('45^o Rotation')
ylabel(''); xlabel('')

subplot(2,3,4)
scatter(nodes4(:,1),nodes4(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(['Monte Carlo, ',num2str(S),' evaluations']); xlabel('')

subplot(2,3,5)
scatter(nodes5(:,1),nodes5(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(''); xlabel('')

subplot(2,3,6)
scatter(nodes6(:,1),nodes6(:,2),'.black')
hold on;
ezcontour(@(x1,x2) mvnpdf([x1,x2],mu,Omega), dom*1.5,100)
hold off;
axis([dom,dom])
title('')
ylabel(''); xlabel('')
